"""
Collection of tests for fields.
"""
import numpy as np
import pytest

from quvac.field.dipole import DipoleAnalytic, create_multibeam
from quvac.field.external_field import ExternalField, ProbePumpField
from quvac.field.gaussian import GaussianAnalytic
from quvac.field.maxwell import MaxwellMultiple
from quvac.field.model import EBInhomogeneity
from quvac.field.utils import convert_tau, get_intensity, get_max_edge_amplitude
from quvac.grid import setup_grids

##########################################################################################
# GAUSSIAN FIELD
##########################################################################################

def get_default_gaussian_params(field_mode="maxwell"):
    return {
        "field_type": f"paraxial_gaussian_{field_mode}",
        "focus_x": [0.,0.,0.],
        "focus_t": 0.,
        "theta": 0,
        "phi": 0,
        "beta": 0,
        "lam": 800e-9,
        "w0": 4*800e-9,
        "tau": 25e-15,
        "W": 25,
        "phase0": 0,
        "order": 0,
    }


def get_default_grid_params():
    return {
        "mode": "dynamic",
        "geometry": "z",
        "transverse_factor": 10,
        "longitudinal_factor": 6,
        "time_factor": 2,
    }


def get_gaussian(order=0):
    gauss_params = get_default_gaussian_params("maxwell")
    gauss_params["order"] = order
    grid_params = get_default_grid_params()
    grid, _ = setup_grids([gauss_params], grid_params)
    field = GaussianAnalytic(gauss_params, grid)
    return field


def check_fields(E, B, check_dtype=True, mode="real"):
    for idx in range(3):
        condition = E[idx].shape == B[idx].shape
        assert condition, "Electric and magnetic field array shapes should be equal."
    if check_dtype:
        for idx in range(3):
            dtype = np.float64 if mode == "real" else np.complex128
            condition = E[idx].dtype == dtype
            assert condition, "Field should have appropriate data type."


@pytest.mark.parametrize("order, mode", [(0,"real"), (5,"real"), (0,"complex")])
def test_gaussian(order, mode):
    gaussian = get_gaussian(order)
    E,B = gaussian.calculate_field(t=0., mode=mode)
    check_fields(E, B, mode=mode)


@pytest.mark.parametrize("polarization", ["linear", "right-circular", "left-circular"])
def test_circular_polarization(polarization):
    gauss_params = get_default_gaussian_params("maxwell")
    gauss_params["polarization"] = polarization
    grid_params = get_default_grid_params()
    grid, _ = setup_grids([gauss_params], grid_params)
    field = GaussianAnalytic(gauss_params, grid)
    E,B = field.calculate_field(t=0.)
    check_fields(E, B)


def get_default_grid(params):
    grid_params = get_default_grid_params()
    if params[0]["field_type"].startswith("dipole"):
        grid_params["transverse_factor"] = 6
    grid, _ = setup_grids(params, grid_params)
    grid.get_k_grid()
    return grid


@pytest.fixture
def gaussian_maxwell():
    params = [get_default_gaussian_params("maxwell")]
    grid = get_default_grid(params)
    field = ExternalField(params, grid)
    return field


def test_gaussian_maxwell(gaussian_maxwell):
    E,B = gaussian_maxwell.calculate_field(t=0.)
    check_fields(E, B, check_dtype=False)


def test_incorrect_field_type_for_maxwell():
    params = [get_default_gaussian_params("incorrect_type")]
    grid = get_default_grid(params)
    with pytest.raises(NotImplementedError):
        ExternalField(params, grid)
    
    with pytest.raises(NotImplementedError):
        MaxwellMultiple(params, grid)


def test_zero_energy_field():
    params = get_default_gaussian_params("maxwell")
    params["W"] = 0
    grid = get_default_grid([params])
    field = MaxwellMultiple([params], grid)
    condition = np.allclose(field.a1, 0.) and np.allclose(field.a2, 0.)
    assert condition, "Spectral coefficients should be zero for W=0"


def get_two_default_gaussians():
    g1 = get_default_gaussian_params("maxwell")
    g2 = get_default_gaussian_params("analytic")
    g2["theta"] = 180
    return [g1, g2]


def test_pump_probe():
    fields = get_two_default_gaussians()
    grid = get_default_grid(fields)
    field = ProbePumpField(fields, grid)
    field.calculate_field(t=0.)


##########################################################################################
# DIPOLE FIELD
##########################################################################################


def get_default_dipole(field_mode="maxwell", envelope="gauss", dipole_type="electric"):
    return {
        "field_type": f"dipole_{field_mode}",
        "envelope": envelope,
        "dipole_type": dipole_type,
        "focus_x": [0.,0.,0.],
        "focus_t": 0.,
        "theta": 0,
        "phi": 0,
        "lam": 800e-9,
        "tau": 25e-15,
        "W": 25,
    }


@pytest.mark.parametrize("envelope, dipole_type",
                         [("gauss", "electric"),("plane", "magnetic")])
def test_dipole(envelope, dipole_type):
    params = get_default_dipole(envelope=envelope, dipole_type=dipole_type)
    grid = get_default_grid([params])
    field = DipoleAnalytic(params, grid)
    E,B = field.calculate_field(t=0.)
    check_fields(E, B, mode="real")


def test_dipole_incorrect_envelope():
    params = get_default_dipole(envelope="incorrect_type")
    grid = get_default_grid([params])
    with pytest.raises(NotImplementedError):
        DipoleAnalytic(params, grid)


@pytest.mark.parametrize("n_beams, multibeam_mode",
                         [(4,"belt"), (6,"belt"), 
                          (12,"sphere")])
def test_multibeam(n_beams, multibeam_mode):
    # create seed beam
    field_seed = get_default_gaussian_params()
    field_seed.update({
        "theta": 90,
        "phi": 0,
    })

    if multibeam_mode == "belt":
        phi0 = 360/(n_beams*2)
    elif multibeam_mode == "sphere":
        phi0 = 360/(n_beams//3*2)

    multibeam_params = create_multibeam(field_seed, n_beams=n_beams,
                                        mode=multibeam_mode, phi0=phi0)
    assert len(multibeam_params) == n_beams, f"Multibeam should consist of {n_beams} "
    f"beams but it has {len(multibeam_params)}"


##########################################################################################
# MODEL FIELD
##########################################################################################


def get_inhom_params(field_inhom, envelope):
    return {
        "field_type": "eb_inhomogeneity",
        "field_inhom": field_inhom,
        "envelope_type": envelope,
        "theta": 90,
        "phi": 90,
        "W": 1,
        "w0": 2*800e-9,
        "lam": 800e-9,
    }


@pytest.mark.parametrize("field_inhom, envelope",
                         [("magnetic","gauss"), ("electric","gauss-modulated")])
def test_model_fields(field_inhom, envelope):
    inhom_params = get_inhom_params(field_inhom, envelope)
    gauss_params = get_default_gaussian_params()
    grid = get_default_grid([gauss_params])

    field = EBInhomogeneity(inhom_params, grid)
    E,B = field.calculate_field(t=0.)
    check_fields(E, B, mode="real")


def test_incorrect_model_fields():
    gauss_params = get_default_gaussian_params()
    grid = get_default_grid([gauss_params])

    inhom_params_1 = get_inhom_params("incorrect_type", "gauss")
    inhom_params_2 = get_inhom_params("electric", "incorrect_type")
    for inhom_params in [inhom_params_1, inhom_params_2]:
        with pytest.raises(NotImplementedError):
            EBInhomogeneity(inhom_params, grid)


##########################################################################################
# FIELD UTILS
##########################################################################################


def test_convert_tau():
    # it is our reference
    tau = 25
    taus_ref = convert_tau(tau, mode="1/e^2")
    # check with other conventions
    modes = ["FWHM", "FWHM-Intensity", "std"]
    for mode in modes:
        taus = convert_tau(taus_ref[mode], mode=mode)
        condition = all(np.isclose(t1,t2) for t1,t2 
                        in zip(taus_ref.values(), taus.values()))  # noqa: B905
        assert condition, "Results for different conventions should be the same."

    # check with incorrect convention name
    with pytest.raises(NotImplementedError):
        convert_tau(tau, mode="incorrect_type")


def test_intensity(gaussian_maxwell):
    intensity = get_intensity(gaussian_maxwell, t=0.)
    assert np.all(intensity > 0), "Intensity should be >0."

    E,_ = gaussian_maxwell.calculate_field(t=0.)
    amplitude_edge = get_max_edge_amplitude(E)
    assert np.all(amplitude_edge > 0), "Field amplitude on edges should be >0."




"""
Useful generic utilities.
"""

import importlib
import inspect
import os
from pathlib import Path
import pkgutil
import platform
import resource
import shutil

import numpy as np
import pyfftw
import yaml


def read_yaml(yaml_file):
    """
    Read a YAML file and return its contents.

    Parameters
    ----------
    yaml_file : str
        Path to the YAML file.

    Returns
    -------
    dict
        Contents of the YAML file.
    """
    with open(yaml_file) as stream:
        try:
            data = yaml.safe_load(stream)
            return data
        except yaml.YAMLError as exc:
            print(exc)
            return exc


def write_yaml(yaml_file, data):
    """
    Write data to a YAML file.

    Parameters
    ----------
    yaml_file : str
        Path to the YAML file.
    data : dict
        Data to write to the YAML file.
    """
    with open(yaml_file, "w") as outfile:
        yaml.dump(data, outfile, default_flow_style=False)


def format_time(seconds):
    """
    Format time in seconds to a human-readable string.

    Parameters
    ----------
    seconds : float
        Time in seconds.

    Returns
    -------
    str
        Formatted time string.
    """
    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    out_str = [
        f"{days:.0f} days" * bool(days),
        f"{hours:.0f} h" * bool(hours),
        f"{minutes:.0f} min" * bool(minutes),
        f"{seconds:.2f} s",
    ]
    return " ".join(out_str)


def format_memory(mem):
    """
    Format memory in kilobytes to a human-readable string.

    Parameters
    ----------
    mem : float
        Memory in kilobytes.

    Returns
    -------
    str
        Formatted memory string.
    """
    units = "KB MB GB TB".split()
    idx = 0
    while mem > 1024:
        mem /= 1024
        idx += 1
    return f"{mem:.2f} {units[idx]}"


def save_wisdom(ini_file, wisdom_file=None, add_host_name=False):
    """
    Save FFTW wisdom to a file.

    Parameters
    ----------
    ini_file : str
        Path to the initialization file.
    wisdom_file : str, optional
        Path to save the FFTW wisdom, by default None.
    add_host_name : bool, optional
        Whether to add the host name to the wisdom file name, by default False.
    """
    if wisdom_file is None:
        wisdom_path = os.path.dirname(ini_file)
        if not os.path.exists(wisdom_path):
            Path(wisdom_path).mkdir(parents=True, exist_ok=True)
        wisdom_name = "fftw-wisdom"
        if add_host_name:
            wisdom_name += platform.node()
        wisdom_file = os.path.join(wisdom_path, wisdom_name)
    else:
        wisdom_path = os.path.dirname(wisdom_file)
        if not os.path.exists(wisdom_path):
            Path(wisdom_path).mkdir(parents=True, exist_ok=True)
    wisdom = pyfftw.export_wisdom()
    with open(wisdom_file, "wb") as f:
        f.write(b"\n".join(wisdom))


def load_wisdom(wisdom_file):
    """
    Load FFTW wisdom from a file.

    Parameters
    ----------
    wisdom_file : str
        Path to the FFTW wisdom file.

    Returns
    -------
    tuple
        FFTW wisdom.
    """
    with open(wisdom_file, "rb") as f:
        wisdom = f.read()
    return tuple(wisdom.split(b"\n"))


def get_maxrss():
    """
    Get the maximum resident set size used (in kilobytes).

    Returns
    -------
    int
        Maximum resident set size used.
    """
    return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss


def zip_directory_shutil(directory_path, output_path):
    """
    Zip a directory using shutil.

    Parameters
    ----------
    directory_path : str
        Path to the directory to zip.
    output_path : str
        Path to save the zipped directory.
    """
    shutil.make_archive(output_path, 'zip', directory_path)


def find_classes_in_package(package_name):
    """
    Find all class names in a given package.

    Parameters
    ----------
    package_name : str
        Name of the package.

    Returns
    -------
    list of str
        List of class names in the package.
    """
    classes = []
    
    # Import the package
    package = importlib.import_module(package_name)
    
    # Recursively find all modules in the package
    for _, module_name, _ in pkgutil.walk_packages(package.__path__, 
                                                   package.__name__ + "."):
        try:
            module = importlib.import_module(module_name)
            
            # Inspect module members and find classes
            for name, obj in inspect.getmembers(module, inspect.isclass):
                # Ensure class belongs to the module (not an imported one)
                if obj.__module__ == module_name:
                    classes.append(f"{module_name}.{name}")
        
        except Exception as e:
            print(f"Skipping {module_name} due to error: {e}")

    return classes


def round_to_n(x, n):
    return round(x, -int(np.floor(np.sign(x) * np.log10(abs(x)))) + n)



#!/usr/bin/env python3
"""
Script to launch Vacuum Emission simulation, do postprocessing
and measure performance.

Usage:

.. code-block:: bash

    python simulation.py -i <input>.yaml -o <output_dir> --wisdom <wisdom_file>
"""

import argparse
import logging
import os
from pathlib import Path
import time

import numexpr as ne
import pyfftw

from quvac import config
from quvac.field.external_field import ExternalField, ProbePumpField
from quvac.grid import setup_grids
from quvac.integrator.vacuum_emission import VacuumEmission
from quvac.log import (
    get_grid_params,
    get_performance_stats,
    get_postprocess_info,
    get_postprocess_stats,
    get_test_timings,
    simulation_end_str,
    simulation_start_str,
)
from quvac.postprocess import VacuumEmissionAnalyzer
from quvac.utils import get_maxrss, load_wisdom, read_yaml, save_wisdom, write_yaml

_logger = logging.getLogger("simulation")


def parse_args(description="Calculate quantum vacuum signal for external fields."):
    """
    Parse command-line arguments.

    Returns
    -------
    argparse.Namespace
        Parsed command-line arguments.
    """
    argparser = argparse.ArgumentParser(description=description)
    argparser.add_argument(
        "--input", "-i", default=None, help="Input yaml file with field and grid params"
    )
    argparser.add_argument(
        "--output", "-o", default=None, help="Path to save simulation data to"
    )
    argparser.add_argument(
        "--wisdom", default="wisdom/fftw-wisdom", help="File to save pyfftw-wisdom"
    )
    return argparser.parse_args()


def check_dirs(ini_file, save_path):
    """
    Check if directories exist and create them if necessary.

    Parameters
    ----------
    ini_file : str
        Path to the initialization file.
    save_path : str
        Path to save the results.

    Returns
    -------
    save_path : str
        Validated save path.
    """
    assert os.path.isfile(ini_file), f"{ini_file} is not a file or does not exist"
    if save_path is None:
        save_path = os.path.dirname(ini_file)
    if not os.path.exists(save_path):
        Path(save_path).mkdir(parents=True, exist_ok=True)
    return save_path


def get_filenames(ini_file, save_path, wisdom_file):
    """
    Get filenames for saving simulation data.

    Parameters
    ----------
    ini_file : str
        Path to the initialization file.
    save_path : str
        Path to save the results.
    wisdom_file : str
        Path to save pyfftw-wisdom.

    Returns
    -------
    files : dict
        Dictionary containing filenames for saving simulation data.
    """
    ini_config = read_yaml(ini_file)
    mode = ini_config.get('mode', 'simulation_postprocess')
    files = {}
    files['save_path'] = save_path
    files['ini'] = ini_file
    files['wisdom'] = wisdom_file
    files['amplitudes'] = os.path.join(save_path, "amplitudes.npz")
    files['spectra'] = os.path.join(save_path, "spectra.npz")
    files['performance'] = os.path.join(save_path, f"{mode}_performance.yml")
    files['logger'] = os.path.join(save_path, f"{mode}.log")
    return files


def get_dirs(ini_file, save_path, wisdom_file):
    """
    Get directories for saving simulation data.

    Parameters
    ----------
    ini_file : str
        Path to the initialization file.
    save_path : str
        Path to save the results.
    wisdom_file : str
        Path to save pyfftw-wisdom.

    Returns
    -------
    files : dict
        Dictionary containing directories for saving simulation data.
    """
    save_path = check_dirs(ini_file, save_path)
    files = get_filenames(ini_file, save_path, wisdom_file)
    return files


def set_precision(precision):
    """
    Set global precision for calculations.

    Parameters
    ----------
    precision : str
        Precision type, either "float32" or "float64".
    """
    if precision == "float32":
        config.FDTYPE = "float32"
        config.CDTYPE = "complex64"
    else:
        config.FDTYPE = "float64"
        config.CDTYPE = "complex128"


def run_simulation(ini_config, fields_params, files, timings, memory):
    """
    Run the vacuum emission simulation.

    Parameters
    ----------
    ini_config : dict
        Dictionary containing the initialization configuration.
    fields_params : list of dict
        List of dictionaries containing the field parameters.
    files : dict
        Dictionary containing filenames for saving simulation data.
    timings : dict
        Dictionary to store timing information.
    memory : dict
        Dictionary to store memory usage information.

    Returns
    -------
    timings : dict
        Updated dictionary with timing information.
    memory : dict
        Updated dictionary with memory usage information.
    """
    grid_params = ini_config["grid"]
    perf_params = ini_config.get("performance", {})

    # Determine integrator type
    integrator_params = ini_config.get("integrator", {})
    integrator_type = integrator_params.get("type", "vacuum_emission")
    channels = integrator_type.endswith("channels")
    if channels:
        probe_pump_idx = integrator_params.get("probe_pump_idx", None)

    # Set up number of threads
    nthreads = perf_params.get("nthreads", os.cpu_count())
    ne.set_num_threads(nthreads)
    pyfftw_threads = perf_params.get("pyfftw_threads", nthreads)
    use_wisdom = perf_params.get("use_wisdom", True)

    # Check if it's a test run to plan resources
    test_run = perf_params.get("test_run", False)
    test_timesteps = perf_params.get("test_timesteps", 5)

    # Load fftw-wisdom if possible
    if use_wisdom and os.path.exists(files['wisdom']):
        pyfftw.import_wisdom(load_wisdom(files['wisdom']))

    # Get grids
    grid_xyz, grid_t = setup_grids(fields_params, grid_params)
    grid_xyz.get_k_grid()
    grid_print = get_grid_params(grid_xyz, grid_t)
    _logger.info(grid_print)
    _logger.info("MILESTONE: Grids are created\n")

    # Shorten time grid for the test run
    if test_run:
        expected_timesteps = len(grid_t)
        grid_t = grid_t[:test_timesteps]
        _logger.info(f"Performing test run for {test_timesteps} timesteps\n")

    # Field setup
    _logger.info(
        "Field constructor:\n" "===================================================="
    )
    if not channels:
        field = ExternalField(fields_params, grid_xyz, nthreads=pyfftw_threads)
    else:
        field = ProbePumpField(
            fields_params,
            grid_xyz,
            probe_pump_idx=probe_pump_idx,
            nthreads=pyfftw_threads,
        )
    timings['field_setup'] = time.perf_counter()
    _logger.info("====================================================\n")
    _logger.info("MILESTONE: Fields are set up")

    # Calculate amplitudes
    if not channels:
        log_message = "Calculating vacuum emission amplitude with external field..."
    else:
        probe_pump = field.probe_pump_idx
        log_message = (
            "Calculating vacuum emission amplitude for probe channel...\n"
            f"    Probe idx: {probe_pump['probe']}\n"
            f"    Pump  idx: {probe_pump['pump']}"
        )
    _logger.info(log_message)
    vacem = VacuumEmission(field, grid_xyz, nthreads=pyfftw_threads, channels=channels)
    timings['vacem_setup'] = time.perf_counter()
    timings['integral'] = vacem.calculate_amplitudes(grid_t, 
                                                     save_path=files['amplitudes'])
    timings['amplitudes'] = time.perf_counter()
    memory['maxrss_amplitudes'] = get_maxrss()
    _logger.info("MILESTONE: Amplitudes are calculated")

    Nt = len(grid_t)
    timings['per_iteration'] = (timings['amplitudes'] - timings['field_setup']) / Nt

    if test_run:
        test_run_str_print = get_test_timings(timings, len(grid_t), expected_timesteps)
        _logger.info(test_run_str_print)
        print(test_run_str_print)
    return timings, memory


def postprocess_simulation(ini_config, files, fields_params):
    """
    Perform postprocessing on the simulation data.

    Parameters
    ----------
    ini_config : dict
        Dictionary containing the initialization configuration.
    files : dict
        Dictionary containing filenames for saving simulation data.
    fields_params : list of dict
        List of dictionaries containing the field parameters.
    """
    # Get postprocess params from ini config
    postprocess_params = ini_config.get("postprocess", {})

    kwargs = {
        "perp_type": postprocess_params.get("perp_polarization_type", None),
        "perp_field_idx": postprocess_params.get("perp_field_idx", 1),
        "stokes": postprocess_params.get("stokes", False),
        "calculate_xyz_background": postprocess_params.get("calculate_xyz_background",
                                                           False),
        "bgr_idx": postprocess_params.get("bgr_idx", False),
        "calculate_spherical": postprocess_params.get("calculate_spherical", False),
        "spherical_params": postprocess_params.get("spherical_params", {}),
        "calculate_discernible": postprocess_params.get("calculate_discernible", False),
        "discernibility": postprocess_params.get("discernibility", "angular"),
        "add_signal_bg": postprocess_params.get("add_signal_bg", False),
    }

    # Do postprocessing
    postprocess_print = get_postprocess_info(postprocess_params)
    _logger.info(postprocess_print)

    modes = postprocess_params.get("modes", ["total", "polarization"])
    if kwargs["perp_type"] is None:
         modes = ["total"]
    for mode in modes:
        save_path = files['spectra'].replace(".npz", f"_{mode}.npz")
        analyzer = VacuumEmissionAnalyzer(
            fields_params, data_path=files['amplitudes'], save_path=save_path
        )
        analyzer.get_spectra(
            mode=mode,
            **kwargs
        )
    _logger.info("MILESTONE: Spectra are calculated from amplitudes")


def quvac_simulation(ini_file, save_path=None, wisdom_file="wisdom/fftw-wisdom"):
    """
    Launch a single quvac simulation for given <ini>.yaml file.

    Parameters
    ----------
    ini_file : str
        Path to the initialization file.
    save_path : str, optional
        Path to save simulation results to, by default None.
    wisdom_file : str, optional
        Path to save pyfftw-wisdom, by default "wisdom/fftw-wisdom".
    """
    # Check that ini file and save_path exist
    files = get_dirs(ini_file, save_path, wisdom_file)
    
    # Load and parse ini yaml file
    ini_config = read_yaml(ini_file)
    # One can choose either to perform just simulation, just postprocess or both
    mode = ini_config.get('mode', 'simulation_postprocess')
    do_simulation = 'simulation' in mode
    do_postprocess = 'postprocess' in mode

    # Setup logger
    logging.basicConfig(
        filename=files['logger'],
        filemode="w",
        encoding="utf-8",
        level=logging.DEBUG,
        format="%(message)s",
    )

    # Start time
    time_log_start = time.asctime(time.localtime())
    start_print = simulation_start_str.format(time_log_start)
    _logger.info(start_print)
    timings = {}
    timings['start'] = time.perf_counter()
    memory = {'maxrss_amplitudes': 0}

    # Set up global precision for calculations
    perf_params = ini_config.get("performance", {})
    precision = perf_params.get("precision", "float64")
    set_precision(precision)
    _logger.info(f"Using {precision} precision")

    fields_params = ini_config["fields"]
    if isinstance(fields_params, dict):
        fields_params = list(fields_params.values())

    if do_simulation:
        timings, memory = run_simulation(ini_config, fields_params, 
                                         files, timings, memory)
    # Calculate spectra
    if do_postprocess:
        postprocess_simulation(ini_config, files, fields_params)
    timings['postprocess'] = time.perf_counter()
    timings['total'] = timings['postprocess'] - timings['start']

    # Save gained wisdom (for fftw)
    save_wisdom(ini_file, wisdom_file)

    memory['maxrss_total'] = get_maxrss()
    perf_stats = {"timings": timings, "memory": memory}

    write_yaml(files['performance'], perf_stats)

    if do_simulation:
        perf_print = get_performance_stats(perf_stats)
        print(perf_print)
        _logger.info(perf_print)
    elif do_postprocess:
        perf_print = get_postprocess_stats(perf_stats)
        print(perf_print)
        _logger.info(perf_print)

    print("Simulation finished!")

    # End time
    time_log_end = time.asctime(time.localtime())
    end_print = simulation_end_str.format(time_log_end)
    _logger.info(end_print)


def main_simulation():
    """
    Main function to run the simulation script.
    """
    args = parse_args()
    quvac_simulation(args.input, args.output, args.wisdom)


if __name__ == "__main__":
    main_simulation()



"""
Functions for creating spatial and temporal grids.

1. ``GridXYZ`` class that calculates the spatial grid and its 
Fourier counterpart.

2. Helper function ``setup_grids`` that automatically creates 
grids either dynamically or from given grid sizes.

.. warning::
    The dynamic grid creation is experimental and should be used with caution.
    Especially, when testing on new physical problems, it is recommended to
    visualize participating Maxwell fields and check the grid sizes.

.. note::
    Parts of dynamical grid creation for gaussian beams were initially
    implemented by Alexander Blinne (``get_kmax_from_bw``, ``get_xyz_size``,
    ``get_t_size``).
"""

from collections.abc import Iterable
from copy import deepcopy

import numexpr as ne
import numpy as np
import pyfftw
from scipy.constants import c, pi

from quvac import config


class GridXYZ:
    """
    Calculates spatial grid and its Fourier counterpart.

    Parameters
    ----------
    grid : tuple of np.array
        A tuple containing three numpy arrays (x, y, z) representing the spatial grid 
        for field discretization.
    
    Attributes
    ----------
    grid : tuple of np.array
        The spatial grid for field discretization.
    grid_shape : tuple of int
        The shape of the grid.
    xyz : tuple of np.array
        The meshgrid of the spatial grid.
    dxyz : tuple of float
        The grid spacing in each dimension.
    dV : float
        The volume element of the grid.
    kgrid : tuple of np.array
        The k-space grid.
    kgrid_shifted : tuple of np.array
        The shifted k-space grid.
    dkxkykz : list of float
        The spacing in k-space for each dimension.
    dVk : float
        The volume element in k-space.
    kmeshgrid : tuple of np.array
        The meshgrid of the k-space grid.
    kabs : np.array
        The absolute value of the k-vectors.
    e1x, e1y, e1z : np.array
        The x, y, and z components of the first polarization vector.
    e2x, e2y, e2z : np.array
        The x, y, and z components of the second polarization vector.
    """

    def __init__(self, grid):
        # Define spatial grid
        self.grid = grid
        self.grid_shape = tuple(ax.size for ax in grid)
        self.xyz = self.x, self.y, self.z = np.meshgrid(
            *grid, indexing="ij", sparse=True
        )
        self.dxyz = tuple(ax[1] - ax[0] for ax in grid)
        self.dV = np.prod(self.dxyz)
        self.kgrid = None

    def get_k_grid(self):
        """
        Calculate the k-space grid and polarization vectors.

        Returns
        -------
        None
        """
        if self.kgrid is None:
            self._calculate_k_grid()

    def _calculate_k_grid(self):
        for i, ax in enumerate("xyz"):
            Nx, dx = self.grid_shape[i], self.dxyz[i]
            k = (2 * pi * np.fft.fftfreq(Nx, dx)).astype(config.FDTYPE)
            setattr(self, f"k{ax}", k)

        self.kgrid = (self.kx, self.ky, self.kz)
        self.kgrid_shifted = tuple(
            np.fft.fftshift(k) for k in (self.kx, self.ky, self.kz)
        )
        self.dkxkykz = [ax[1] - ax[0] for ax in self.kgrid]
        self.dVk = np.prod(self.dkxkykz)

        self.kmeshgrid = np.meshgrid(
            self.kx, self.ky, self.kz, indexing="ij", sparse=True
        )
        kx, ky, kz = self.kmeshgrid
        self.kabs = kabs = ne.evaluate("sqrt(kx**2 + ky**2 + kz**2)") # noqa: F841
        kperp = ne.evaluate("sqrt(kx**2 + ky**2)") # noqa: F841

        # Polarization vectors
        self.e1x = ne.evaluate(
            "where((kx==0) & (ky==0), 2*(kz>0)-1, kx * kz / (kperp*kabs))"
        )
        self.e1y = ne.evaluate("where((kx==0) & (ky==0), 0, ky * kz / (kperp*kabs))")
        self.e1z = ne.evaluate("where((kx==0) & (ky==0), 0, -kperp / kabs)")

        self.e2x = ne.evaluate("where((kx==0) & (ky==0), 0, -ky / kperp)")
        self.e2y = ne.evaluate("where((kx==0) & (ky==0), 1, kx / kperp)")
        # self.e2y = ne.evaluate("where((kx==0) & (ky==0), 2*(kz>0)-1, kx / kperp)")
        self.e2z = 0


def get_ek(theta, phi):
    """
    Calculate k-vector for given spherical angles theta and phi.

    Parameters
    ----------
    theta : float
        The polar angle in radians.
    phi : float
        The azimuthal angle in radians.

    Returns
    -------
    numpy.ndarray
        A 3-element array representing the k-vector.
    """
    ek = np.array(
        [np.sin(theta) * np.cos(phi), np.sin(theta) * np.sin(phi), np.cos(theta)]
    )
    return ek


def get_pol_basis(theta, phi):
    """
    Calculate polarization basis for given spherical angles theta and phi.

    Parameters
    ----------
    theta : float
        The polar angle in radians.
    phi : float
        The azimuthal angle in radians.

    Returns
    -------
    tuple of numpy.ndarray
        A tuple containing two 3-element arrays representing the polarization
        basis vectors e1 and e2.
    """
    e1 = np.array(
        [np.cos(theta) * np.cos(phi), np.cos(theta) * np.sin(phi), -np.sin(theta)],
        dtype=config.FDTYPE
    )
    e2 = np.array([-np.sin(phi), np.cos(phi), 0], dtype=config.FDTYPE)
    return e1, e2


def get_polarization_vector(theta, phi, beta):
    """
    Calculate polarization vector for given spherical angles theta and phi
    and polarization angle beta.

    Parameters
    ----------
    theta : float
        The polar angle in radians.
    phi : float
        The azimuthal angle in radians.
    beta : float
        The polarization angle in radians.

    Returns
    -------
    numpy.ndarray
        3-element array representing the polarization vector.
    """
    e1, e2 = get_pol_basis(theta, phi)
    ebeta = e1*np.cos(beta) + e2*np.sin(beta)
    return ebeta


def gaussian_bandwidth(field_params):
    """
    Calculate the gaussian bandwidth.

    Parameters
    ----------
    field_params : dict
        Dictionary containing the field parameters. Required keys are:
            - tau : float
                Pulse duration.
            - w0 : float, optional
                Beam waist. If 'w0' is not provided, 'w0x' and 'w0y' are used.

    Returns
    -------
    tuple of float
        A tuple containing the perpendicular and longitudinal bandwidths 
        kbw_perp, kbw_long).

    Notes
    -----
    The bandwidths are calculated from the Fourier transform of transverse and 
    longitudinal Gaussian profiles:

    - exp(-r**2/w0**2) -> exp(-w0**2*kperp**2/4) -> kperp_bw ~ 2*2/w0
    
    - exp(-t**2/(tau/2)**2) -> exp(-tau**2*k**2/16) -> k_long ~ 2*4/(c*tau)
    """
    tau = field_params["tau"]

    # gaussian might have circular or elliptic cross section
    if "w0" in field_params:
        w0 = field_params["w0"]
    elif "w0x" in field_params:
        w0 = min(field_params["w0x"], field_params["w0y"])

    kbw_perp = 4 / w0
    kbw_long = 8 / (c * tau)

    return kbw_perp, kbw_long


def dipole_bandwidth(field_params):
    """
    Calculate the dipole bandwidth.

    Parameters
    ----------
    field_params : dict
        Dictionary containing the field parameters. Required keys are:
            - lam : float
                Wavelength of the pulse.
            - tau : float
                Pulse duration.

    Returns
    -------
    tuple of float
        A tuple containing the perpendicular and longitudinal bandwidths 
        (kbw_perp, kbw_long).

    Notes
    -----
    The length scales are taken from:
    I. Gonoskov et al. "Dipole pulse theory: Maximizing the field amplitude 
    from 4π focused laser pulses." PRA 86.5 (2012): 053836.

    l_para = 0.58 * lam
    
    l_perp = 0.4 * lam
    """
    lam, tau = [field_params[k] for k in "lam tau".split()]
    
    l_long = 0.58*lam
    l_perp = 0.4*lam
    tau_bw = 8 / (c * tau)

    # Add up bandwidth coming from time and length scales
    kbw_perp = 4 / l_perp + tau_bw
    kbw_long = 4 / l_long + tau_bw

    return kbw_perp, kbw_long


def get_bw(field_params):
    """
    Calculate the bandwidths for a given field type.

    Parameters
    ----------
    field_params : dict
        Dictionary containing the field parameters.

    Returns
    -------
    tuple of float
        A tuple containing the perpendicular and longitudinal bandwidths 
        (kbw_perp, kbw_long).

    Raises
    ------
    NotImplementedError
        If the field type is not supported.
    """
    ftype = field_params["field_type"]
    if "gaussian" in ftype:
        assert "w0" in field_params or "w0x" in field_params, (
            "Gaussian field parameters must have either 'w0' or 'w0x' key"
        )
        bws = gaussian_bandwidth(field_params)
    elif "dipole" in ftype:
        bws = dipole_bandwidth(field_params)
    return bws


def _check_keys(field_params, required_keys):
    err_msg = f"Field parameters must have {required_keys} keys"
    assert all(key in field_params for key in required_keys), err_msg


def get_kmax_from_bw(field_params):
    """
    Calculate the maximum k-vector from bandwidth along each dimension 
    for a given field.

    Parameters
    ----------
    field_params : dict
        Dictionary containing the field parameters. Required keys are:
            - lam : float
                Wavelength of the pulse.
            - tau : float
                Pulse duration.
            - theta : float
                Polar angle in degrees.
            - phi : float
                Azimuthal angle in degrees.

    Returns
    -------
    numpy.ndarray
        A 3-element array representing the maximum k-vector along each dimension.

    Notes
    -----
    The maximum k-vector is calculated based on the bandwidths derived from the 
    field parameters.
    """
    required_keys = "lam tau theta phi".split()
    _check_keys(field_params, required_keys)

    lam, _, theta, phi = [field_params[k] for k in required_keys]

    k = 2 * np.pi / lam
    theta *= pi / 180
    phi = phi * pi / 180 if np.sin(theta) != 0.0 else 0.0

    ek = get_ek(theta, phi)
    e1, e2 = get_pol_basis(theta, phi)

    kbw_perp, kbw_long = get_bw(field_params)

    k0 = ek * k
    kmax = np.abs(k0 + ek * kbw_long)

    for beta in np.linspace(0, 2 * np.pi, 64, endpoint=False):
        kp = k0 + ek * kbw_long + kbw_perp * (np.cos(beta) * e1 + np.sin(beta) * e2)
        kmax = np.maximum(kmax, np.abs(kp))
    
    return kmax


def get_kmax_grid(field_params):
    """
    Calculate the maximum k-vector along each dimension for a given field type.

    Parameters
    ----------
    field_params : dict
        Dictionary containing the field parameters.

    Returns
    -------
    numpy.ndarray
        A 3-element array representing the maximum k-vector along each dimension.

    Raises
    ------
    NotImplementedError
        If the field type is not supported.
    """
    ftype = field_params["field_type"]
    if ("gaussian" in ftype) or ("dipole" in ftype):
        kmax = get_kmax_from_bw(field_params)
    else:
        raise NotImplementedError(f"{ftype} field type is not supported")
    return kmax


def get_xyz_size(fields, box_size, grid_res=1, equal_resolution=False):
    """
    Calculate necessary spatial resolution.

    Parameters
    ----------
    fields : dict or list of dict
        Parameters of participating fields.
    box_size : sequence of float
        Box size for 3 dimensions.
    grid_res : float, optional
        Controls the resolution (default is 1).
    equal_resolution : bool, optional
        Flag indicating if equal resolution in every dimension is needed 
        (default is False).

    Returns
    -------
    list of int
        List containing the number of grid points along each dimension.
    """
    if isinstance(fields, dict):
        fields = list(fields.values())
    box_size = np.array(box_size)

    kmax = np.zeros((3,))
    for field_params in fields:
        kmax = np.maximum(kmax, get_kmax_grid(field_params))

    # if a square box is required
    if equal_resolution:
        kmax = np.max(kmax) * np.ones(3)

    N = grid_res * np.ceil(box_size * 3 * kmax / pi)
    N = [pyfftw.next_fast_len(int(n)) for n in N]
    return N


def get_t_size(t_start, t_end, lam, grid_res=1):
    """
    Calculate necessary temporal resolution.

    Parameters
    ----------
    t_start : float
        Start time.
    t_end : float
        End time.
    lam : float
        Wavelength.
    grid_res : float, optional
        Factor to scale the resolution (default is 1).

    Returns
    -------
    float
        Number of time steps.
    """
    fmax = c / lam
    return int(np.ceil((t_end - t_start) * fmax * 6 * grid_res))


def get_max_duration(fields_params):
    tau_max = 0
    for field in fields_params:
        tau = field.get("tau", 0)
        # spectral chirp increases effective pulse duration
        if field["field_type"] == "paraxial_gaussian_spectral_maxwell":
            alpha = field.get("alpha_chirp", 0)
            tau *= np.sqrt(1 + alpha**2)
        tau_max = max([tau_max, tau])
    return tau_max


def get_box_size(fields_params, grid_params):
    """
    Calculate the necessary box size for the spatial grid.

    Parameters
    ----------
    fields_params : list of dict
        Parameters of participating fields. Each dictionary should contain:
            - field_type : str
                Type of the field.
            - w0 : float, optional
                Beam waist for focused fields.
            - tau : float, optional
                Pulse duration.
    grid_params : dict
        Grid parameters. Required keys are:
            - transverse_factor : float
                Factor to scale the transverse size.
            - longitudinal_factor : float
                Factor to scale the longitudinal size.

    Returns
    -------
    tuple of float
        A tuple containing the transverse and longitudinal sizes.

    Notes
    -----
    The box size is calculated based on the maximum beam waist and pulse duration
    among the fields, scaled by the transverse and longitudinal factors.
    """
    perp_max = 0
    for field in fields_params:
        ftype = field["field_type"]
        if "gauss" in ftype:
            w0, w0x, w0y = [field.get(key, 0) for key in ("w0", "w0x", "w0y")]
            length = length = np.max([w0, w0x, w0y])
            # length = field.get("w0", 0)
        elif "dipole" in ftype:
            length = c * field.get("tau", 0) / 4
        else:
            length = 0
        perp_max = np.maximum(length, perp_max)
    
    # tau_max = max([field.get("tau", 0) for field in fields_params])
    tau_max = get_max_duration(fields_params)

    transverse_size = perp_max * grid_params["transverse_factor"]
    longitudinal_size = tau_max * c * grid_params["longitudinal_factor"]

    return float(transverse_size), float(longitudinal_size)


def create_dynamic_grid(fields_params, grid_params):
    """
    Dynamically create grids from given laser parameters.
    One should be careful with this option.

    Parameters
    ----------
    fields_params : list of dict
        Parameters of participating fields. 
    grid_params : dict
        Grid parameters.

    Returns
    -------
    dict
        Updated grid parameters including calculated box size, number of spatial points,
        and number of temporal points.
    """
    grid_params_upd = deepcopy(grid_params)
    # Create spatial box
    collision_geometry = grid_params.get("collision_geometry", "z")

    tau_max = max([field.get("tau", 0) for field in fields_params])
    lam_min = min([field.get("lam", 1e10) for field in fields_params])

    transverse_size, longitudinal_size = get_box_size(fields_params, grid_params)

    box_xyz = [0, 0, 0]
    for i, ax in enumerate("xyz"):
        box_xyz[i] = longitudinal_size if ax in collision_geometry else transverse_size
    grid_params_upd["box_xyz"] = box_xyz

    # Number of spatial pts
    box_size = np.array(box_xyz) / 2
    Nxyz_ = get_xyz_size(fields_params, box_size)
    res = grid_params.get("spatial_resolution", 1)
    if isinstance(res, Iterable):
        assert len(res) == 3, "Spatial resolution must be a list of 3 values or a "
        "single value"
        Nxyz = [int(Nx * res) for Nx, res in zip(Nxyz_, res, strict=True)]
    elif type(res) in (int, float):
        Nxyz = [int(Nx * res) for Nx in Nxyz_]
    grid_params_upd["Nxyz"] = Nxyz

    # Create temporal box
    t0 = tau_max * grid_params["time_factor"]
    Nt = get_t_size(-t0 / 2, t0 / 2, lam_min)

    # Number of temporal pts
    grid_params_upd["box_t"] = t0
    grid_params_upd["Nt"] = Nt * grid_params.get("time_resolution", 1)
    return grid_params_upd


def setup_grids(fields_params, grid_params):
    """
    Create spatial and temporal grids from given sizes or
    dynamically from field parameters (e.g., duration and waist size).

    Parameters
    ----------
    fields_params : list of dict
        Parameters of participating fields. Each dictionary should contain:
            - field_type : str
                Type of the field.
            - lam : float
                Wavelength of the pulse.
            - w0 : float, optional
                Beam waist for focused fields.
            - tau : float, optional
                Pulse duration.

    grid_params : dict
        Grid parameters. Required keys are:
            -  mode : str
                Mode of grid creation ('dynamic' or 'static').
        Keys for 'static' mode:
            - box_xyz : tuple of float
                Box size for the spatial grid.
            - Nxyz : tuple of int
                Number of grid points along each spatial dimension.
            - Nt : int
                Number of temporal points.
            - box_t : float or tuple of float
                Time duration or start and end times for the temporal grid.
        Keys for 'dynamic' mode:
            - collision_geometry : str
                Specifies the collision geometry ('x', 'y', 'z').
            - transverse_factor : float
                Factor to scale the transverse size.
            - longitudinal_factor : float
                Factor to scale the longitudinal size.
            - time_factor : float
                Factor to scale the time duration.
            - spatial_resolution : float or list of float, optional
                Controls the spatial resolution.
            - time_resolution : float, optional
                Controls the temporal resolution.
            - ignore_idx : list of int, optional
                Indices of fields to ignore for dynamic grid creation.

    Returns
    -------
    tuple
        A tuple containing:
            - grid_xyz : quvac.grid.GridXYZ
                The spatial grid object.
            - grid_t : numpy.ndarray
                The temporal grid array.

    Notes
    -----
    The spatial and temporal grids are created based on the mode specified in the grid 
    parameters. If 'dynamic' mode is selected, the grids are created dynamically based 
    on the field parameters.
    """
    if grid_params["mode"] == "dynamic":
        if isinstance(fields_params, dict):
            fields_params = list(fields_params.values())
        # filter out fields that should not contribute to dynamic grid creation
        ignore_idx = grid_params.get("ignore_idx", None)
        if ignore_idx is not None:
            fields_params = [field for idx,field in enumerate(fields_params) 
                            if idx not in ignore_idx]
        grid_params = create_dynamic_grid(fields_params, grid_params)

    x0, y0, z0 = grid_params["box_xyz"]
    Nx, Ny, Nz = grid_params["Nxyz"]
    x = np.linspace(-0.5 * x0, 0.5 * x0, Nx, endpoint=Nx % 2, dtype=config.FDTYPE)
    y = np.linspace(-0.5 * y0, 0.5 * y0, Ny, endpoint=Ny % 2, dtype=config.FDTYPE)
    z = np.linspace(-0.5 * z0, 0.5 * z0, Nz, endpoint=Nz % 2, dtype=config.FDTYPE)
    grid_xyz = GridXYZ((x, y, z))

    Nt = grid_params["Nt"]
    box_t = grid_params["box_t"]
    # Allow non-symmetric time-grids
    if isinstance(box_t, float):
        Nt += int(1 - Nt % 2)
        t0 = box_t
        grid_t = np.linspace(
            -0.5 * t0, 0.5 * t0, Nt, endpoint=True, dtype=config.FDTYPE
        )
    else:
        t_start, t_end = box_t
        grid_t = np.linspace(t_start, t_end, Nt, endpoint=True, dtype=config.FDTYPE)
    return grid_xyz, grid_t



"""
Configuration file for simulations. 

``FDTYPE`` and ``CDTYPE`` control the float precision for the 
calculations (``float32`` or ``float64``).

``DEFAULT_SUBMITIT_PARAMS`` is a dictionary with default parameters for
job submission to the cluster.
"""

DEFAULT_SUBMITIT_PARAMS = {
    "slurm_partition": "hij-gpu",
    "cpus_per_task": 16,
    "slurm_mem": "20GB",
    "timeout_min": 240,
}

FDTYPE = "float64"
CDTYPE = "complex128"



"""
Collection of tests for simulations.
"""
import os
from pathlib import Path

import numpy as np
import pytest

from quvac.postprocess import (
    get_simulation_fields,
    get_spectra_from_Stokes,
    integrate_spherical,
    signal_in_detector,
)
from quvac.simulation import quvac_simulation
from quvac.simulation_parallel import quvac_simulation_parallel
from quvac.utils import read_yaml, write_yaml
from tests.config_for_tests import DEFAULT_CONFIG_PATH

##########################################################################################
# SIMULATION
##########################################################################################


@pytest.fixture(scope="session")
def get_tmp_path(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("simulation")
    return tmp_path


def save_ini(path, ini_data):
    path = os.path.join(str(path), "test_simulation")
    Path(path).mkdir(parents=True, exist_ok=True)

    ini_file = os.path.join(path, "ini.yml")
    write_yaml(ini_file, ini_data)
    return ini_file


def test_simulation(tmp_path):
    ini_data = read_yaml(DEFAULT_CONFIG_PATH)
    ini_file = save_ini(tmp_path, ini_data)
    quvac_simulation(ini_file)

    ini_data["mode"] = "postprocess"
    ini_file = save_ini(tmp_path, ini_data)
    quvac_simulation(ini_file)


def test_simulation_test(tmp_path):
    ini_data = read_yaml(DEFAULT_CONFIG_PATH)
    ini_data["performance"]["test_run"] = True
    ini_file = save_ini(tmp_path, ini_data)
    quvac_simulation(ini_file)


def test_channels(tmp_path):
    ini_data = read_yaml(DEFAULT_CONFIG_PATH)
    ini_data["integrator"]["type"] = "vacuum_emission_channels"
    ini_file = save_ini(tmp_path, ini_data)
    quvac_simulation(ini_file)


def test_precision(tmp_path):
    ini_data = read_yaml(DEFAULT_CONFIG_PATH)
    ini_data["performance"]["precision"] = "float32"
    ini_data["postprocess"]["perp_polarization_type"] = None
    ini_file = save_ini(tmp_path, ini_data)
    quvac_simulation(ini_file)


##########################################################################################
# POSTPROCESS
##########################################################################################


def test_postprocess(tmp_path):
    ini_data = read_yaml(DEFAULT_CONFIG_PATH)
    # test spherical and discernible signal
    ini_data["postprocess"].update({
        "calculate_spherical": True,
        "calculate_discernible": True,
        "perp_polarization_type": "local axis",
    })
    ini_file = save_ini(tmp_path, ini_data)
    quvac_simulation(ini_file)

    # test stokes, xyz_background
    ini_data["mode"] = "postprocess"
    ini_data["postprocess"].update({
        "calculate_xyz_background": True,
        "bgr_idx": 1,
        "stokes": True,
    })
    ini_file = save_ini(tmp_path, ini_data)
    quvac_simulation(ini_file)

    # test loading stokes parameters
    folder = os.path.dirname(ini_file)
    data = np.load(os.path.join(folder, "spectra_total.npz"))
    P0 = data["N_xyz"]
    data_p = np.load(os.path.join(folder, "spectra_polarization.npz"))
    P1, P2, P3 = [data_p[key] for key in "P1 P2 P3".split()]
    for basis in "linear linear-45 circular".split():
        Nf, Np = get_spectra_from_Stokes(P0, P1, P2, P3, basis=basis)
        assert np.allclose(Nf + Np, P0), "Total signal should be equal to the summ of"
        " perp polarization modes."

    # test loading fields from ini_file
    fields = get_simulation_fields(ini_file)
    assert len(fields) == 2, "There should be 2 fields in the default simulation."

    # test signal in detector
    N_sph, k, theta, phi = [data[key] for key in "N_sph k theta phi".split()]
    N_angular = integrate_spherical(N_sph, (k,theta,phi), axs_integrate=['k'])
    detector = {
        "theta0": 0,
        "phi0": 0,
        "dtheta": 15,
        "dphi": 15,
    }
    N_detector = signal_in_detector(N_angular, theta, phi, detector, align_to_max=True)
    assert N_detector > 0, "Signal should be positive"


def test_mix_bg_and_signal(tmp_path):
    ini_data = read_yaml(DEFAULT_CONFIG_PATH)
    ini_data["postprocess"].update({
        "modes": ["mix_signal_bg"],
        "add_signal_bg": True,
    })
    ini_file = save_ini(tmp_path, ini_data)
    quvac_simulation(ini_file)


##########################################################################################
# SIMULATION PARALLEL
##########################################################################################


def test_parallel_simulation(tmp_path):
    # run usual simulation
    ini_data = read_yaml(DEFAULT_CONFIG_PATH)
    ini_file = save_ini(tmp_path, ini_data)
    quvac_simulation(ini_file)
    
    folder = os.path.dirname(ini_file)
    data = np.load(os.path.join(folder, "spectra_total.npz"))
    N_total = data["N_total"]

    ini_data["performance"]["nthreads"] = 2
    ini_file = save_ini(tmp_path, ini_data)
    quvac_simulation_parallel(ini_file)
    data = np.load(os.path.join(folder, "spectra_total.npz"))
    N_total_parallel = data["N_total"]

    assert np.isclose(N_total, N_total_parallel), "Sequential and parallel results "
    "should be the same."





"""
Various field profiles and interfaces to unite them.

.. note::
    After implementing a new field profile, it should be added to the 
    ``ANALYTIC_FIELDS`` and ``SPATIAL_MODEL_FIELDS`` dicts in the 
    ``quvac.field.__init__``. These dictionaries connect ``field_type``
    keywords in the ``ini.yaml`` files and particular field implementations.
"""
from quvac.field.dipole import DipoleAnalytic
from quvac.field.gaussian import (
    GaussianAnalytic,
    GaussianEllipticAnalytic,
    GaussianSpectral,
)
from quvac.field.laguerre_gaussian import LaguerreGaussianAnalytic
from quvac.field.model import EBInhomogeneity
from quvac.field.planewave import PlaneWave

ANALYTIC_FIELDS = {
    "dipole_analytic": DipoleAnalytic,
    "paraxial_gaussian_analytic": GaussianAnalytic,
    "elliptic_gaussian_analytic": GaussianEllipticAnalytic,
    "laguerre_gaussian_analytic": LaguerreGaussianAnalytic,
    "eb_inhomogeneity": EBInhomogeneity,
    "plane_wave": PlaneWave,
}

SPATIAL_MODEL_FIELDS = {
    "dipole_maxwell": DipoleAnalytic,
    "paraxial_gaussian_maxwell": GaussianAnalytic,
    "elliptic_gaussian_maxwell": GaussianEllipticAnalytic,
    "paraxial_gaussian_spectral_maxwell": GaussianSpectral,
    "laguerre_gaussian_maxwell": LaguerreGaussianAnalytic,
}


#!/usr/bin/env python3
"""
Script to run Bayesian optimization on cluster with Slurm.

Optimization parameters (optimized variables, objectives and constraints)
should be located in `ini.yml` at key `optimization`.

Usage:

.. code-block:: bash

    optimization.py -i <input>.yaml -o <output_dir>
"""
from copy import deepcopy
import os
from pathlib import Path
import time
import warnings

from ax.modelbridge.generation_strategy import GenerationStep, GenerationStrategy
from ax.modelbridge.registry import Models
from ax.service.ax_client import AxClient, ObjectiveProperties
import numpy as np
from submitit import AutoExecutor, DebugJob, LocalJob

from quvac.cluster.config import DEFAULT_SUBMITIT_PARAMS
from quvac.postprocess import integrate_spherical, signal_in_detector
from quvac.simulation import parse_args, quvac_simulation
from quvac.utils import read_yaml, round_to_n, write_yaml


def prepare_params_for_ax(params, ini_file):
    """
    Prepare parameters for Ax optimization.

    Parameters
    ----------
    params : dict
        Dictionary containing parameter categories and their bounds.
    ini_file : str
        Path to the initial configuration file.

    Returns
    -------
    list of dict
        List of parameter descriptions formatted for Ax optimization.

    Note
    ----
    Only field parameters could be optimized. They are given as dictionaries
    {"field_1": {"param_1": range_1, "param_2": range_2}, ...} which are transformed
    to ["field_1:param_1", "field_1:param_2", ...] parameter names.

    Also the path to default `ini.yml` is passed. For every optimization trial it is 
    loaded, optimized parameters are changed and the simulation is submitted.
    """
    params_ax = []
    for category_key, category in params.items():
        for key, val in category.items():
            loc_name = f"{category_key}:{key}"
            param_descr = {
                "name": loc_name,
                "type": "range",
                "bounds": val,
                "value_type": "float",
            }
            params_ax.append(param_descr)
    params_ax.append({"name": "ini_default", "type": "fixed", "value": ini_file})
    return params_ax


def objective_signal_in_detector(data, obj_params):
    """
    Calculate the signal detected within a specified detector region.

    Parameters
    ----------
    data : dict
        Dictionary containing simulation results, including spherical grid data.
    obj_params : dict
        Dictionary containing detector parameters.

    Returns
    -------
    float
        The signal detected within the specified detector region.
    """
    detectors = obj_params["detectors"]
    detectors = [detectors] if isinstance(detectors, dict) else detectors
    k, theta, phi, N_sph = [data[key] for key in "k theta phi N_sph".split()]
    N_angular = integrate_spherical(N_sph, (k,theta,phi), axs_integrate=['k'])

    N_detector = 0
    for detector in detectors:
        N_detector += signal_in_detector(N_angular, theta, phi, detector,
                                         align_to_max=False)
    return N_detector


def objective_discernible_signal_in_detector(data, obj_params):
    """
    Calculate the discernible signal detected within a specified detector region.

    Parameters
    ----------
    data : dict
        Dictionary containing simulation results, including spherical grid data.
    obj_params : dict
        Dictionary containing detector parameters.

    Returns
    -------
    float
        The signal detected within the specified detector region.
    """
    detectors = obj_params["detectors"]
    detectors = [detectors] if isinstance(detectors, dict) else detectors
    k, theta, phi, N_sph, discernible = [
        data[key] for key in "k theta phi N_sph discernible".split()
    ]
    N_angular = integrate_spherical(N_sph, (k,theta,phi), axs_integrate=['k'])

    N_detector = 0
    for detector in detectors:
        N_detector += signal_in_detector(N_angular*discernible, theta, phi, detector,
                                         align_to_max=False)
    return N_detector


def update_energies(ini_data, energy_params):
    """
    Update the energy distribution among fields for optimization.

    Parameters
    ----------
    ini_data : dict
        Dictionary containing the initial configuration data.
    energy_params : dict
        Dictionary containing the energy parameters to optimize.

    Returns
    -------
    dict
        Updated configuration data with modified energy distribution.

    Raises
    ------
    AssertionError
        If the number of fields does not match the number of optimized parameters plus 
        one.
    """
    ini = deepcopy(ini_data)
    optimization_params = ini["optimization"]
    energy_fields = optimization_params["energy_fields"]
    fields, opt_fields = [energy_fields[key] for key
                          in "fields optimized_fields".split()]
    err_msg = ("While optimizing energy distribution, it is required to have "
               "exactly one more field than free parameters")
    assert len(fields) == len(opt_fields)+1, err_msg

    optimization_params = ini_data["optimization"]
    scales = optimization_params.get("scales", {})
    scale = scales.get("W", 1)

    W_total = 0.
    for param_key, param in energy_params.items():
        category, key = param_key.split(":")
        ini["fields"][category][key] = float(param * scale)
        W_total += param
    
    # fix energy of remaining field
    idx_remain = list(set(fields) - set(opt_fields))[0]
    W_remain = np.maximum(1. - W_total, 0.)
    ini["fields"][f"field_{idx_remain}"]["W"] = float(W_remain * scale)
    return ini


def collect_metrics(data, obj_params, metric_names=("N_total")):
    """
    Collect metrics from simulation results.

    Parameters
    ----------
    data : dict
        Dictionary containing simulation results.
    obj_params : dict
        Dictionary containing objective parameters.
    metric_names : list of str, optional
        List of metric names to collect. Default is ("N_total").

    Returns
    -------
    dict
        Dictionary containing the collected metrics.
    """
    N_disc = data.get("N_disc", 0)
    N_total = data.get("N_total", 0)

    metrics = {}
    metrics["N_total"] = (float(N_total), 0.0)
    if N_disc is not None:
        metrics["N_disc"] = (float(N_disc), 0.0)

    if "detectors" in obj_params:
        N_detector = objective_signal_in_detector(data, obj_params)
        metrics["N_detector"] = (float(N_detector), 0.0)
        N_detector_disc = objective_discernible_signal_in_detector(data, obj_params)
        metrics["N_detector_disc"] = (float(N_detector_disc), 0.0)

    # filter metrics
    metrics = {k:v for k,v in metrics.items() if k in metric_names}
    return metrics


def quvac_evaluation(params, metric_names=("N_total")):
    """
    Evaluate a single trial of the quvac simulation.

    Parameters
    ----------
    params : dict
        Dictionary containing trial parameters.
    metric_names : list of str, optional
        List of metric names to evaluate. Default is ("N_total").

    Returns
    -------
    dict
        Dictionary containing metrics such as `N_disc`, `N_total`, and optionally 
        `N_detector`.
    """
    ini_file = params["ini_default"]
    ini_data = read_yaml(ini_file)
    params.pop("ini_default")
    trial_idx = params.pop("trial_idx")

    optimization_params = ini_data["optimization"]
    obj_params = optimization_params.get("objectives_params", {})
    scales = optimization_params.get("scales", {})

    # Create ini.yml file for current trial
    trial_str = str(trial_idx).zfill(3)
    save_folder = ini_data.get("save_path", os.path.dirname(ini_file))
    save_path = os.path.join(save_folder, trial_str)

    energy_params = {}
    # Update_parameters for current trial
    for param_key, param in params.items():
        category, key = param_key.split(":")
        if key != "W":
            scale = scales.get(key, 1)
            ini_data["fields"][category][key] = float(param * scale)
        else:
            energy_params[param_key] = param

    # treat separately energy distribution parameters
    if energy_params:     
        ini_data = update_energies(ini_data, energy_params)

    # Save ini file
    ini_path = os.path.join(save_path, "ini.yml")
    Path(os.path.dirname(ini_path)).mkdir(parents=True, exist_ok=True)
    write_yaml(ini_path, ini_data)

    # Run simulation
    quvac_simulation(ini_path)

    # Load results
    data = np.load(os.path.join(save_path, "spectra_total.npz"))
    metrics = collect_metrics(data, obj_params, metric_names)
    return metrics


def check_energy_constraint(trial_index_to_param):
    """
    Check if the total energy budget constraint is satisfied.

    Parameters
    ----------
    trial_index_to_param : dict
        Dictionary mapping trial indices to parameter dictionaries. Each parameter 
        dictionary contains the energy distribution among fields.

    Returns
    -------
    bool
        True if the total energy budget constraint is satisfied for all trials, False 
        otherwise.

    Raises
    ------
    Warning
        If the total energy budget constraint is violated for any trial.
    """
    energy_ok = True
    for _, params in trial_index_to_param.items():
        energies = []
        for param_key, param in params.items():
            if param_key != "ini_default":
                category, key = param_key.split(":")
                if key == "W":
                    energies.append(param)
        
        W_total = np.sum(energies)
        eps = 1.0 - W_total
        if eps < 0 and not np.isclose(abs(eps), 0.0, atol=1e-5):
            warnings.warn("Fixed total energy budget constraint is violated! "
                          "Probably, optimization fails to find new prospective points "
                          "and is stuck in local minima.", stacklevel=2)
            energy_ok = False
            break
    return energy_ok


def check_repeated_samples(trial_index_to_param, last_samples, fail_count, patience=3):
    """
    Check for repeated samples in the optimization process.

    Parameters
    ----------
    trial_index_to_param : dict
        Dictionary mapping trial indices to parameter dictionaries.
    last_samples : list of tuple
        List of parameter tuples from the most recent trials.
    fail_count : int
        Counter for the number of consecutive repeated samples.
    patience : int, optional
        Maximum number of consecutive repeated samples allowed before stopping. Default 
        is 3.

    Returns
    -------
    tuple
        A tuple containing:
        - continue_sampling (bool): Whether to continue sampling new trials.
        - last_samples (list of tuple): Updated list of parameter tuples.
        - fail_count (int): Updated counter for repeated samples.

    Notes
    -----
    If the number of repeated samples exceeds the `patience` value, a warning is issued,
    and sampling is stopped.
    """
    continue_sampling = True
    trials = deepcopy(trial_index_to_param)
    for _, params in trials.items():
        params.pop("ini_default")
        # round-up ints and floats for comparison
        params_list = []
        for k,v in tuple(sorted(params.items())):
            if (isinstance(v, int) or isinstance(v, float)) and not np.isclose(v, 0.0):
                # we use 5 significant digits
                v = round_to_n(v,5)
            params_list.append((k,v))
        params_tuple = tuple(params_list)

        if last_samples and params_tuple == last_samples[-1]:
            fail_count += 1
            warnings.warn(f"Trial {len(last_samples)-1} is identical to previous: "
                          f"{params}. Fail count: {fail_count}", stacklevel=2)
        else:
            fail_count = 0
        
        last_samples.append(params_tuple)

        if fail_count >= patience:
            warnings.warn(f"Number of repeated samples exceeded patience ({patience} "
                          "tries)!", stacklevel=2)
            continue_sampling = False
    return continue_sampling, fail_count


def check_sampled_trials(trial_index_to_param, last_samples, fail_count):
    """
    Check if the sampled trials are valid.

    Parameters
    ----------
    trial_index_to_param : dict
        Dictionary mapping trial indices to parameter dictionaries.
    last_samples : list of tuple
        List of parameter tuples from the most recent trials.
    fail_count : int
        Counter for the number of consecutive repeated samples.

    Returns
    -------
    tuple
        A tuple containing:
        - continue_optimization (bool): Whether to continue the optimization process.
        - last_samples (list of tuple): Updated list of parameter tuples.
        - fail_count (int): Updated counter for repeated samples.

    Notes
    -----
    This function checks two conditions:
    1. Whether the total energy budget constraint is satisfied.
    2. Whether there are repeated samples in the optimization process.
    If either condition fails, the optimization process is terminated.
    """
    energy_ok = check_energy_constraint(trial_index_to_param)
    continue_sampling, fail_count = check_repeated_samples(trial_index_to_param, 
                                                           last_samples, fail_count)

    continue_optimization = energy_ok and continue_sampling
    return continue_optimization, fail_count


def run_optimization(ax_client, executor, n_trials, max_parallel_jobs, experiment_file,
                     metric_names=("N_total")):
    """
    Run Bayesian optimization using Ax and Submitit.

    Parameters
    ----------
    ax_client : ax.service.ax_client.AxClient
        Ax client for managing the optimization process.
    executor : submitit.AutoExecutor
        Submitit executor for running jobs on a cluster.
    n_trials : int
        Total number of trials to run.
    max_parallel_jobs : int
        Maximum number of parallel jobs to run.
    experiment_file : str
        Path to save the Ax experiment data.
    metric_names : list of str, optional
        List of metric names to evaluate. Default is ("N_total").

    Returns
    -------
    None
    """
    jobs = []
    submitted_jobs = 0
    # variables for early optimization stopping
    continue_optimization = True
    last_samples = []
    fail_count = 0
    # Run until all the jobs have finished and our budget is used up.
    while (continue_optimization and submitted_jobs < n_trials) or jobs:
        for job, trial_idx in jobs:
            # Poll if any jobs completed
            # Local and debug jobs don't run until .result() is called.
            if job.done() or type(job) in [LocalJob, DebugJob]:
                result = job.result()
                ax_client.complete_trial(trial_index=trial_idx, raw_data=result)
                jobs.remove((job, trial_idx))
                ax_client.save_to_json_file(experiment_file)

        # sample new points if optimization is not terminated
        if continue_optimization:
            # Schedule new jobs if there is availablity
            trial_index_to_param, _ = ax_client.get_next_trials(
                max_trials=min(max_parallel_jobs - len(jobs), n_trials - submitted_jobs)
            )
            # Check that sampled trials satisfy the requirements
            if trial_index_to_param:
                flags = check_sampled_trials(trial_index_to_param,last_samples,
                                             fail_count)
                continue_optimization, fail_count = flags

                if continue_optimization:
                    for trial_idx, params in trial_index_to_param.items():
                        params["trial_idx"] = trial_idx
                        job = executor.submit(quvac_evaluation, params, metric_names)
                        submitted_jobs += 1
                        jobs.append((job, trial_idx))
                        time.sleep(1)
                else:
                    warnings.warn("Terminating optimization... Finishing last "
                                  "trials...", stacklevel=2)
                

def setup_generation_strategy(num_random_trials=6):
    """
    Setup custom generation strategy.

    Parameters
    ----------
    num_random_trials : int, optional
        Number of random trials to perform before starting Bayesian optimization. 
        Default is 6.

    Returns
    -------
    ax.modelbridge.generation_strategy.GenerationStrategy
        The configured generation strategy.
    """
    gs = GenerationStrategy(
        steps=[
            GenerationStep(model=Models.SOBOL, num_trials=num_random_trials),
            GenerationStep(model=Models.GPEI, num_trials=-1),
        ]
    )
    return gs


def cluster_optimization(ini_file, save_path=None, wisdom_file=None):
    """
    Launch optimization of quvac simulation for a given initial configuration file.

    Parameters
    ----------
    ini_file : str
        Path to the initial configuration file (YAML format).
    save_path : str, optional
        Path to save simulation results. If not provided, defaults to the directory of 
        `ini_file`.
    wisdom_file : str, optional
        Path to save FFTW wisdom. Default is None.

    Returns
    -------
    None
    """
    # Check that ini file and save_path exists
    assert os.path.isfile(ini_file), f"{ini_file} is not a file or does not exist"
    if save_path is None:
        save_path = os.path.dirname(ini_file)
    if not os.path.exists(save_path):
        Path(save_path).mkdir(parents=True, exist_ok=True)
    experiment_file = os.path.join(save_path, "experiment.json")

    ini_default = read_yaml(ini_file)
    optimization_params = ini_default["optimization"]
    cluster_params = optimization_params.get("cluster", {})

    # Check that optimization parameters are only field_parameters
    optimization_keys = list(optimization_params["parameters"].keys())
    err_msg = (
        f"Only field parameters could be optimized but you have {optimization_keys}"
    )
    assert sum([key.startswith("field") for key in optimization_keys]) == len(
        optimization_keys
    ), err_msg

    # Prepare parameters for optimization in ax style
    params_for_ax = prepare_params_for_ax(optimization_params["parameters"], ini_file)

    # custom generation strategy
    gs_params = optimization_params.get("gs_params", {})
    num_random_trials = gs_params.get("num_random_trials", 6)
    gs = setup_generation_strategy(num_random_trials=num_random_trials)

    # Set up optimization client
    ax_client = AxClient(generation_strategy=gs)
    objectives = optimization_params["objectives"]
    track_metrics = optimization_params.get("track_metrics", [])
    # collect metric names to keep track of
    metric_names = [name for name, flag in objectives]
    metric_names += track_metrics

    ax_client.create_experiment(
        name=optimization_params.get("name", "test_optimization"),
        parameters=params_for_ax,
        objectives={
            name: ObjectiveProperties(minimize=flag) for name, flag in objectives
        },
        parameter_constraints=optimization_params.get("parameter_constraints", None),
        outcome_constraints=optimization_params.get(
            "outcome_constraints", None
        ),
        tracking_metric_names=track_metrics,
    )

    # Set up sibmitit AutoExecutor
    cluster = cluster_params.get("cluster", "local")
    sbatch_params = cluster_params.get("sbatch_params", DEFAULT_SUBMITIT_PARAMS)
    max_parallel_jobs = cluster_params.get("max_parallel_jobs", 3)
    log_folder = os.path.join(save_path, "submitit_logs")

    executor = AutoExecutor(folder=log_folder, cluster=cluster)
    if cluster == "slurm":
        executor.update_parameters(slurm_array_parallelism=max_parallel_jobs)
        executor.update_parameters(**sbatch_params)
    elif "timeout_min" in cluster_params:
        executor.update_parameters(timeout_min=cluster_params["timeout_min"])

    n_trials = optimization_params.get("n_trials", 10)

    run_optimization(ax_client, executor, n_trials, max_parallel_jobs, experiment_file,
                     metric_names)
    print("Optimization finished!")


def gather_trials_data(ax_client, metric_names=("N_total", "N_disc")):
    """
    Gather data from completed trials in the Ax experiment.

    Parameters
    ----------
    ax_client : ax.service.ax_client.AxClient
        Ax client managing the optimization process.
    metric_names : list of str, optional
        List of metric names to extract from the trials. Default is 
        ("N_total", "N_disc").

    Returns
    -------
    dict
        Dictionary containing trial parameters and their corresponding metrics.
    """
    metrics = ax_client.experiment.fetch_data().df
    trials = ax_client.experiment.trials
    trials_params = {key: trial.arm.parameters for key, trial in trials.items()
                     if trial.completed_successfully}
    for key in trials_params:
        for metric_name in metric_names:
            condition = (metrics["metric_name"] == metric_name) & (
                metrics["trial_index"] == key
            )
            trials_params[key][metric_name] = metrics.loc[condition]["mean"].iloc[0]
        if "ini_default" in trials_params[key]:
            trials_params[key].pop("ini_default")
    return trials_params


def main_optimization():
    """
    Main function to run the optimization.
    """
    args = parse_args(description="Perform optimization of quvac simulations")
    cluster_optimization(args.input, args.output)


if __name__ == "__main__":
    main_optimization()



#!/usr/bin/env python3
"""
Script to run gridscan simulations on cluster with Slurm.

Gridscan parameters (scanned variables and their grids)
should be located in `ini.yml` at key `variables`.

Usage:

.. code-block:: bash

    gridscan.py -i <input>.yaml -o <output_dir>
"""
from copy import deepcopy
import itertools
import os
from pathlib import Path

import numpy as np
import submitit

from quvac.cluster.config import DEFAULT_SUBMITIT_PARAMS
from quvac.simulation import parse_args, quvac_simulation
from quvac.utils import read_yaml, write_yaml


def _create_grids(variables):
    """
    Create parameter grids for each category and parameter.

    Parameters
    ----------
    variables : dict
        Dictionary containing parameter categories and their bounds. Each parameter
        should be specified as a tuple (start, end, npts).

    Returns
    -------
    dict
        Dictionary containing parameter grids for each category and parameter.
    """
    variables_grid = {}
    for category_key, category in variables.items():
        variables_grid[category_key] = {}
        for param_key, param in category.items():
            start, end, npts = param
            param_grid = list(np.linspace(start, end, npts))
            variables_grid[category_key][param_key] = param_grid
    return variables_grid


def create_parameter_grids(variables):
    """
    Create a grid from (start, end, npts) specified for each parameter.

    Parameters
    ----------
    variables : dict
        Dictionary containing parameter categories and their bounds. The key 'fields'
        is handled separately as it is a 3-level dictionary.

    Returns
    -------
    dict
        Dictionary containing parameter grids for all categories and parameters.
    """
    variables_grid = {}
    fields = variables.get("fields", {})
    if fields:
        fields_grid = _create_grids(fields)
        variables_grid["fields"] = fields_grid
        variables.pop("fields")
    params_grid = _create_grids(variables)
    variables_grid.update(params_grid)
    return variables_grid


def restructure_variables_grid(variables):
    """
    Transform a nested dictionary into a flat dictionary by combining nested keys.

    Parameters
    ----------
    variables : dict
        Dictionary containing parameter grids for each category and parameter.

    Returns
    -------
    tuple
        A tuple containing:
        - param_names : list of list of str
            List of parameter names split into categories and parameter keys.
        - param_grids : list of list
            List of parameter grids corresponding to each parameter.
    """
    variables_grid = {}
    for key, val in variables["fields"].items():
        variables[key] = val
    variables.pop("fields")

    for category_key, category in variables.items():
        for param_key, param in category.items():
            new_key = f"{category_key}:{param_key}"
            variables_grid[new_key] = param

    param_names = list(variables_grid.keys())
    param_names = [param.split(":") for param in param_names]
    param_grids = list(variables_grid.values())
    return param_names, param_grids


def create_ini_files_for_gridscan(ini_default, param_names, param_grids, save_path):
    """
    Create separate `ini.yml` files for every combination of parameters in the 
    grid scan.

    Parameters
    ----------
    ini_default : dict
        Default initial configuration dictionary containing all simulation parameters.
    param_names : list of list of str
        List of parameter names split into categories and parameter keys.
    param_grids : list of list
        List of parameter grids corresponding to each parameter.
    save_path : str
        Path to save the generated `ini.yml` files.

    Returns
    -------
    list of str
        List of file paths to the generated `ini.yml` files.
    """
    ini_files = []
    for parametrization in itertools.product(*param_grids):
        ini_current = deepcopy(ini_default)
        name_local = ""
        for name, param in zip(param_names, parametrization, strict=True):
            category, param_name = name
            if category.startswith("field"):
                ini_current["fields"][category][param_name] = float(param)
            else:
                ini_current[category][param_name] = float(param)

            param = int(param) if np.isclose(param, int(param)) else param
            param_str = str(param) if isinstance(param, int) else f"{param:.2e}"
            param_str = f"{category}:{param_name}_{param_str}"
            name_local = "#".join([name_local, param_str])
        save_path_local = os.path.join(save_path, name_local, "ini.yml")
        Path(os.path.dirname(save_path_local)).mkdir(parents=True, exist_ok=True)
        write_yaml(save_path_local, ini_current)
        ini_files.append(save_path_local)
    return ini_files


def cluster_gridscan(ini_file, save_path=None, wisdom_file=None):
    """
    Launch a grid scan of quvac simulations for a given default `ini.yml` file.

    Parameters
    ----------
    ini_file : str
        Path to the default initial configuration file (YAML format) containing all 
        simulation parameters.
    save_path : str, optional
        Path to save simulation results. If not provided, defaults to the directory of 
        `ini_file`.
    wisdom_file : str, optional
        Path to save FFTW wisdom. Default is None.

    Returns
    -------
    None
    """
    # Check that ini file and save_path exists
    assert os.path.isfile(ini_file), f"{ini_file} is not a file or does not exist"
    if save_path is None:
        save_path = os.path.dirname(ini_file)
    if not os.path.exists(save_path):
        Path(save_path).mkdir(parents=True, exist_ok=True)

    ini_default = read_yaml(ini_file)
    variables = ini_default["variables"]
    cluster_params = variables.get("cluster", {})
    if cluster_params:
        variables.pop("cluster")

    create_grids = variables.get("create_grids", False)
    if "create_grids" in variables:
        variables.pop("create_grids")

    # Create parameter grids if required
    if create_grids:
        variables_grid = create_parameter_grids(variables)
    else:
        variables_grid = variables

    # Restructure variables dict
    param_names, param_grids = restructure_variables_grid(variables_grid)

    # Create yaml files for the grid scan
    ini_files = create_ini_files_for_gridscan(
        ini_default, param_names, param_grids, save_path
    )

    # Set up scheduler
    cluster = cluster_params.get("cluster", "local")
    log_folder = os.path.join(save_path, "submitit_logs")
    sbatch_params = cluster_params.get("sbatch_params", DEFAULT_SUBMITIT_PARAMS)
    max_jobs = cluster_params.get("max_parallel_jobs", 5)
    executor = submitit.AutoExecutor(folder=log_folder, cluster=cluster)
    if cluster == "slurm":
        executor.update_parameters(slurm_array_parallelism=max_jobs)
        executor.update_parameters(**sbatch_params)
    else:
        executor.update_parameters(timeout_min=30)
    print("Submitting jobs...")
    jobs = executor.map_array(quvac_simulation, ini_files)
    print("Jobs submitted, waiting for results...")

    # Wait till all jobs end
    _ = [job.result() for job in jobs]
    print("Grid scan is finished!")


def main_gridscan():
    """
    Main function to run the gridscan.
    """
    args = parse_args(description="Perform gridscan of quvac simulations")
    cluster_gridscan(args.input, args.output)


if __name__ == "__main__":
    main_gridscan()



"""
Analytic expression for dipole wave and function to create multibeam
configuration by compining several focused fields.

----

.. [1] I. Gonoskov et al. "Dipole pulse theory: Maximizing the field
    amplitude from 4 π focused laser pulses." PRA 86.5 (2012): 053836.

.. [2] S. S. Bulanov et al. "Multiple Colliding Electromagnetic Pulses: 
    A Way to Lower the Threshold of e+ e-Pair Production from Vacuum." 
    PRL 104.22 (2010): 220404.
"""
# ruff: noqa: F841
from copy import deepcopy

import numexpr as ne
from scipy.constants import c, pi

from quvac.field.abc import ExplicitField


class DipoleAnalytic(ExplicitField):
    """
    Analytic expression for dipole wave.

    Parameters
    ----------
    field_params : dict
        Dictionary containing the field parameters.
            - focus_x : tuple of float
                Location of spatial focus (x, y, z).
            - focus_t : float
                Location of temporal focus.
            - theta : float
                Spherical angle of dipole moment d0 with z-axis (in degrees).
            - phi : float
                Spherical angle of dipole moment d0 with x-axis (in degrees).
            - lam : float
                Pulse wavelength.
            - tau : float
                Duration.
            - phase0 : float
                Phase delay at focus.
            - W : float, optional
                Energy.
            - envelope : str
                Envelope type, either "plane" or "gauss".
            - dipole_type : str
                Type of dipole, either "electric" or "magnetic".
    grid : quvac.grid.GridXYZ
        Spatial and spectral grid.
    
    Notes
    -----
    Dipole wave expression is from [1]_.

    d0 is along ez by default.
    """
    def __init__(self, field_params, grid):
        super().__init__(field_params, grid)
        
        self.beta = getattr(self, "beta", 0)
        self.envelope = getattr(self, "envelope", "plane")
        self.dipole_type = getattr(self, "dipole_type", "electric")
        self.phase0 = getattr(self, "phase0", 0)

        # Define additional field variables
        self.x0, self.y0, self.z0 = self.focus_x
        self.t0 = self.focus_t
        self.k = 2.0 * pi / self.lam
        self.omega = c * self.k
        self.d0 = 1.

        # Rotate coordinate grid
        self.rotate_coordinates()

        # Define variables independent of time
        self.R_expr = 'sqrt(x**2 + y**2 + z**2)' # radius
        self.R =  ne.evaluate(self.R_expr, global_dict=self.__dict__)

        R_inv_expr = "where(R==0, 0, 1/R)"
        R_inv =  ne.evaluate(R_inv_expr, global_dict=self.__dict__)
        
        self.nx, self.ny, self.nz = [ax*R_inv for ax in (self.x, self.y, self.z)]

        self.EB_dict = {"R": self.R,
                        "c": c,
                        "k": self.k,
                        "omega": self.omega,
                        "phase0": self.phase0,
                        "d0": self.d0,
                        "nx": self.nx,
                        "ny": self.ny,
                        "nz": self.nz}

        self.set_envelope_expressions()
        self.check_energy()

    def set_envelope_expressions(self):
        """
        Defines envelope expressions.
        """
        if self.envelope == "plane":
            env = "1j*exp(-1j*omega*t - 1j*phase0)"
            self.g_expr = env
            self.gdot_expr = f"-1j*omega*{env}"
            self.gdotdot_expr = f"-omega**2*{env}"
            self.E_R0 = "omega**3"
        elif self.envelope == "gauss":
            a2 = "(1/(tau/2)**2)"
            env = f"-1j*exp(-t**2*{a2} - 1j*omega*t - 1j*phase0)"
            self.g_expr = env
            self.gdot_expr = f'{env} * (-2*t*{a2} - 1j*omega)'
            self.gdotdot_expr = (f'{env} * (4*t**2*{a2}**2 - 2*{a2} - omega**2 + '
                                 f'4j*t*omega*{a2})')
            self.E_R0 = (f"{env} * (-2*t*{a2}*(4*t**2*{a2}**2 - 3*omega**2 - 6*{a2})"
                         f" - 1j*omega*(12*t**2*{a2}**2 - omega**2 - 6*{a2}))")
        else:
            raise NotImplementedError(f"Envelope {self.envelope} not implemented")

    def check_energy(self):
        """
        Checks and adjusts the energy of the field.
        """
        self._check_energy()

        if self.modify_energy:
            self.d0 *= self.W_correction

    def _g(self, t):
        return ne.evaluate(self.g_expr, global_dict=self.__dict__)
    
    def _gdot(self, t):
        return ne.evaluate(self.gdot_expr, global_dict=self.__dict__)
    
    def _gdotdot(self, t):
        return ne.evaluate(self.gdotdot_expr, global_dict=self.__dict__)
    
    def _g_plusminus(self, t, sign=1):
        return self._g(t-self.R/c) + sign*self._g(t+self.R/c)
    
    def _gdot_plusminus(self, t, sign=1):
        return self._gdot(t-self.R/c) + sign*self._gdot(t+self.R/c)
    
    def _gdotdot_plusminus(self, t, sign=1):
        return self._gdotdot(t-self.R/c) + sign*self._gdotdot(t+self.R/c)

    def _fix_singularity(self, t):
        # fix divergence at R=0
        Nx,Ny,Nz = self.Ex.shape
        self.Ex[Nx//2,Ny//2,Nz//2] = 0.
        self.Ey[Nx//2,Ny//2,Nz//2] = 0.
        self.Ez[Nx//2,Ny//2,Nz//2] = 4/(3*c**3)*ne.evaluate(self.E_R0,
                                                            global_dict=self.__dict__)

        self.Bx[Nx//2,Ny//2,Nz//2] = 0.
        self.By[Nx//2,Ny//2,Nz//2] = 0.
    
    def calculate_field(self, t, E_out=None, B_out=None, mode="real"):
        """
        Calculates the electric and magnetic fields at a given time step.
        """
        gdotdot_p = self._gdotdot_plusminus(t)
        gdotdot_m = self._gdotdot_plusminus(t, sign=-1)
        gdot_p = self._gdot_plusminus(t)
        gdot_m = self._gdot_plusminus(t, sign=-1)
        g_m = self._g_plusminus(t, sign=-1)

        Bt = ne.evaluate("gdotdot_p/(R*c**2) + gdot_m/(R**2*c)", 
                         global_dict=self.EB_dict)
        Et = ne.evaluate("gdot_p/(c*R**2) + g_m/R**3", global_dict=self.EB_dict)
        
        self.Ex = ne.evaluate('nx*nz*gdotdot_m/(R*c**2) + 3*nx*nz*Et',
                              global_dict=self.EB_dict)
        self.Ey = ne.evaluate('ny*nz*gdotdot_m/(R*c**2) + 3*ny*nz*Et', 
                              global_dict=self.EB_dict)
        self.Ez = ne.evaluate('-(nx**2+ny**2)*gdotdot_m/(R*c**2) + (3*nz**2-1)*Et',
                               global_dict=self.EB_dict)

        self.Bx = ne.evaluate("-ny*Bt", global_dict=self.EB_dict)
        self.By = ne.evaluate("nx*Bt", global_dict=self.EB_dict)
        self.Bz = 0.
        
        # fix divergence at R=0
        self._fix_singularity(t)

        # h-dipole (magnetic) transformation
        if self.dipole_type == "magnetic":
            self.Ex, self.Bx = -self.Bx, self.Ex
            self.Ey, self.By = -self.By, self.Ey
            self.Ez, self.Bz = -self.Bz, self.Ez

        for field in "Ex Ey Ez Bx By Bz".split():
            self.__dict__[field] *= self.d0

        if mode == "real":
            self.convert_fields_to_real()

        E_out, B_out = self.rotate_fields_back(E_out, B_out, mode)
        return E_out, B_out


def create_multibeam(params, n_beams=6, mode='belt', phi0=0, idx0=1):
    """
    Create multibeam configuration from several focused pulses to 
    approximate the dipole wave and achieve higher intensity 
    at focus.

    Parameters
    ----------
    params : dict
        Dictionary containing the field parameters.
    n_beams : int, optional
        Number of beams, by default 6.
    mode : str, optional
        Configuration mode, either 'belt' or 'sphere', by default 'belt'.
    phi0 : float, optional
        Initial angle for the beams, by default 0.
    idx0 : int, optional
        Offset index for `beams` dictionary.

    Returns
    -------
    beams : dict
        Dictionary containing the parameters for each beam.

    Notes
    -----
    Configuration follows from [2]_.
    """
    # distribute the energy
    W_per_beam = params['W'] / n_beams
    beams = {}
    if mode == "sphere":
        n_beams = n_beams // 3
        theta_arr = [0, 45, -45]
    phi_c = 360/n_beams

    for i in range(n_beams):
        params_beam = deepcopy(params)
        params_beam["W"] = W_per_beam
        params_beam["phi"] = i*phi_c + phi0
        params_beam["theta"] = 90
        match mode:
            case "belt":
                beams[f"field_{i+1+idx0}"] = params_beam
            case "sphere":
                for j,theta in enumerate(theta_arr):
                    idx = i*3 + j
                    params_phi = deepcopy(params_beam)
                    params_phi["theta"] += theta
                    beams[f"field_{idx+1+idx0}"] = params_phi
    return beams
